/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2015 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2015 Ryu Takegami (Twitter:@RyuTakegami)
 *
 *	[ Win32API MIDI ]
 */

#ifdef _WIN32
#ifdef MIDI

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include "xm7.h"
#include "device.h"
#include "midi.h"
#include "w32.h"
#include "w32_midi.h"
#include "w32_sch.h"

/*
 *	O[o [N
 */
char szMidiDevice[256];					/* MIDIfoCX */
int nMidiDelay;							/* MIDIx */
BOOL bMidiDelayMode;					/* MIDIx[h */

/*
 *	X^eBbN [N
 */
static HMIDIOUT	hOut = 0;				/* MIDI OUTnh*/
static MIDIHDR	hOutHdr;				/* MIDI OUTwb_*/
static char		szOpenDevice[256];		/* I[ṽfoCX */

static MIDISEND	SendBuf[MAXSENDBUF];	/* MIDIf[^obt@ */
static BYTE		ExclBuf[MAXEXCLBUF];	/* GNXN[Vuf[^obt@ */
static int		nMidiCtrl;				/* MbZ[W̃^Cv*/
static int		nMidiPos;				/* MbZ[W̃f[^JE^ */
static int		nMidiSysCount;			/* RbZ[WXLbvJE^ */
static BOOL		bMidiExcvWait;			/* GNXN[VuMtO */
static DWORD	dwSendReadPtr;			/* MIDIbZ[W  MJE^ */
static DWORD	dwSendWritePtr;			/* MIDIbZ[W  JE^ */
static DWORD	dwExclWritePtr;			/* GNXN[VuJE^ */
static BYTE		nMidiLastData;			/* ŌɑMf[^ */

/*
 *	vg^Cv錾
 */
static void FASTCALL AllNoteOff(void);
static void FASTCALL WaitLastExcvOut(void);


/*
 *	
 */
void FASTCALL InitMIDI(void)
{
	MIDIOUTCAPS moc;
	int nDeviceID;
	UINT i;

	/* [NGA */
	nMidiCtrl = MIDICTRL_READY;
	nMidiLastData = 0x80;
	bMidiExcvWait = FALSE;
	dwSendReadPtr = 0;
	dwSendWritePtr = 0;

	if (!hOut) {
		/* OvfoCX */
		nDeviceID = (int)MIDI_MAPPER;
		for (i=0; i<midiOutGetNumDevs(); i++) {
			midiOutGetDevCaps(i, &moc, sizeof(moc));
			if (!strcmp(moc.szPname, szMidiDevice)) {
				nDeviceID = (int)i;
				break;
			}
		}

		/* MIDI|[gI[v */
		if (nDeviceID >= 0) {
			if (midiOutOpen(&hOut, (UINT)nDeviceID, 0, 0, CALLBACK_NULL)
								== MMSYSERR_NOERROR) {
				AllNoteOff();
				strncpy(szOpenDevice, moc.szPname, sizeof(szOpenDevice));
				return;
			}
		}

		/* I[vs */
		hOut = NULL;
		strncpy(szOpenDevice, "", sizeof(szOpenDevice));
	}
}


/*
 *	N[Abv
 */
void FASTCALL CleanMIDI(void)
{
	if (hOut) {
		AllNoteOff();
		midiOutReset(hOut);
		midiOutClose(hOut);
		hOut = NULL;
	}
}

/*
 *	ZNgԃ`FbN
 */
BOOL FASTCALL SelectCheckMIDI(char *device)
{
	/* gƂĂ|[gƎgp̃|[gvĂ邩`FbN */
	if (strcmp(szOpenDevice, device)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	ZbgfoCXZNg
 */
void FASTCALL midi_reset_notify(void)
{
	/* gƂĂ|[gƎgp̃|[gvĂ邩`FbN */
	if (strcmp(szOpenDevice, szMidiDevice)) {
		/* |[gJĂꍇ͂N[Abv */
		if (hOut) {
			CleanMIDI();
		}

		/* ߂ď */
		InitMIDI();
	}
}


/*
 *	S`lL[It
 */
static void FASTCALL AllNoteOff(void)
{
	DWORD msg;

	if (hOut) {
		WaitLastExcvOut();
		for (msg=0x7bb0; msg<0x7bc0; msg++) {
			midiOutShortMsg(hOut, msg);
		}
	}
}


/*
 *	GNXN[VuMҋ@
 */
static void FASTCALL WaitLastExcvOut(void)
{
	if (bMidiExcvWait) {
		while (midiOutUnprepareHeader(hOut, &hOutHdr,
			sizeof(MIDIHDR)) == MIDIERR_STILLPLAYING);
		bMidiExcvWait = FALSE;
	}
}


/*
 *	GNXN[VuM
 */
static void FASTCALL MidiSendExcv(DWORD ptr, int len)
{
	char buf[1024];
	int i;

	/* f[^Uobt@ɓ] */
	for (i=0; i<len; i++) {
		buf[i] = ExclBuf[(i + ptr) % MAXEXCLBUF];
	}

	/* GNXN[VuM */
	hOutHdr.lpData = (char *)buf;
	hOutHdr.dwFlags = 0;
	hOutHdr.dwBufferLength = len;
	midiOutPrepareHeader(hOut, &hOutHdr, sizeof(MIDIHDR));
	midiOutLongMsg(hOut, &hOutHdr, sizeof(MIDIHDR));

	/* GNXN[VuMtO𗧂Ă */
	bMidiExcvWait = TRUE;
}


/*
 *	f[^ʒm
 */
void FASTCALL midi_notify(BYTE mes)
{
	/* MIDI OUT݂Ȃꍇ͉Ȃ */
	if (!hOut) {
		return;
	}

	/* obt@t̏ꍇ̓f[^̂Ă */
	if (midi_busy) {
		return;
	}

	/* A^CbZ[W */
	switch (mes) {
		case MIDI_TIMING:
		case MIDI_START:
		case MIDI_CONTINUE:
		case MIDI_STOP:
		case MIDI_ACTIVESENSE:
		case MIDI_SYSTEMRESET:
			return;
	}

	/* 1oCg */
	if (nMidiCtrl == MIDICTRL_READY) {
		if (mes & 0x80) {
			SendBuf[dwSendWritePtr].data[0] = mes;
			SendBuf[dwSendWritePtr].time = dwNowTime;
			nMidiPos = 1;

			/* Xe[^XbZ[W */
			switch (mes & 0xf0) {
				case 0x80:		/* m[gIt */
				case 0x90:		/* m[gI */
				case 0xa0:		/* |tHjbNL[vbV */
				case 0xb0:		/* Rg[`FW */
				case 0xe0:		/* sb`x_ */
					nMidiLastData = mes;
					SendBuf[dwSendWritePtr].length = 3;
					nMidiCtrl = MIDICTRL_3BYTES;
					break;

				case 0xc0:		/* vO`FW */
				case 0xd0:		/* `lvbV */
					SendBuf[dwSendWritePtr].length = 2;
					nMidiCtrl = MIDICTRL_2BYTES;
					break;

				default:
					switch (mes) {
						case MIDI_EXCLUSIVE:	/* VXeGNXN[Vu */
							nMidiCtrl = MIDICTRL_EXCLUSIVE;
							SendBuf[dwSendWritePtr].exclptr = dwExclWritePtr;
							ExclBuf[dwExclWritePtr ++] = 0xf0;
							if (dwExclWritePtr >= MAXEXCLBUF) {
								dwExclWritePtr = 0;
							}
							break;
						case MIDI_SONGPOS:		/* \O|WV */
							nMidiCtrl = MIDICTRL_SYSTEM;
							nMidiSysCount = 3;
							break;
						case MIDI_SONGSELECT:	/* \OZNg */
							nMidiCtrl = MIDICTRL_SYSTEM;
							nMidiSysCount = 2;
							break;
						case MIDI_TIMECODE:		/* ^CR[h */
						case MIDI_TUNEREQUEST:	/* `[NGXg */
							nMidiCtrl = MIDICTRL_SYSTEM;
							nMidiSysCount = 1;
							break;
						default:				/* ȊO */
							return;
					}
					break;
			}
		}
		else {
			/* jOXe[^X */
			SendBuf[dwSendWritePtr].length = 3;
			SendBuf[dwSendWritePtr].data[0] = nMidiLastData;
			SendBuf[dwSendWritePtr].data[1] = mes;
			nMidiPos = 2;
			nMidiCtrl = MIDICTRL_3BYTES;
		}
	}
	else {
		/* 2oCgȍ~̏ */
		switch (nMidiCtrl) {
			case MIDICTRL_2BYTES:
			case MIDICTRL_3BYTES:
				SendBuf[dwSendWritePtr].data[nMidiPos] = mes;

				nMidiPos ++;
				if (nMidiPos >= SendBuf[dwSendWritePtr].length) {
					dwSendWritePtr = (WORD)((dwSendWritePtr + 1) % MAXSENDBUF);
					if (dwSendWritePtr == dwSendReadPtr) {
						midi_busy = TRUE;
					}
					nMidiCtrl = MIDICTRL_READY;
				}
				break;

			case MIDICTRL_EXCLUSIVE:
				ExclBuf[dwExclWritePtr] = mes;
				dwExclWritePtr = (WORD)((dwExclWritePtr + 1) % MAXEXCLBUF);

				nMidiPos ++;
				if (mes == MIDI_EOX) {
					/* bZ[WI */
					SendBuf[dwSendWritePtr].length = nMidiPos;
					dwSendWritePtr = (WORD)((dwSendWritePtr + 1) % MAXSENDBUF);
					if (dwSendWritePtr == dwSendReadPtr) {
						midi_busy = TRUE;
					}
					nMidiCtrl = MIDICTRL_READY;
				}
				else if (nMidiPos >= 1024) {
					/* obt@I[o[t[ */
					nMidiCtrl = MIDICTRL_READY;
				}
				break;

			case MIDICTRL_SYSTEM:
				nMidiPos ++;
				if (nMidiPos >= nMidiSysCount) {
					nMidiCtrl = MIDICTRL_READY;
				}
				break;
		}
	}
}

/*
 *	f[^M (x߂)
 */
void FASTCALL MidiSendData(DWORD tick)
{
	DWORD sendmsg;
	DWORD tim;

	for (;;) {
		if (dwSendReadPtr == dwSendWritePtr) {
			/* ŏI݃|CgƈvꍇAI */
			midi_busy = FALSE;
			return;
		}

		/* xԕo߂`FbN */
		tim = SendBuf[dwSendReadPtr].time;
		if (dwNowTime >= tim) {
			if ((dwNowTime - tim) < tick) {
				return;
			}
		}

		/* GNXN[VubZ[W̑MI҂ */
		WaitLastExcvOut();

		if (SendBuf[dwSendReadPtr].data[0] == 0xf0) {
			/* GNXN[VubZ[WM */
			MidiSendExcv(SendBuf[dwSendReadPtr].exclptr,
				SendBuf[dwSendReadPtr].length);
		}
		else {
			/* GNXN[VuȊÕbZ[W𑗐M */
			sendmsg =  SendBuf[dwSendReadPtr].data[0];
			sendmsg |= SendBuf[dwSendReadPtr].data[1] << 8;
			if (SendBuf[dwSendReadPtr].length == 3) {
				sendmsg |= SendBuf[dwSendReadPtr].data[2] << 16;
			}
			midiOutShortMsg(hOut, sendmsg);
		}

		/* JE^XV */
		dwSendReadPtr = (WORD)((dwSendReadPtr + 1) % MAXSENDBUF);
	}
}

#endif	/* MIDI */
#endif	/* _WIN32 */
